import asyncio
import collections
import logging
import signal


class Server:
    def __init__(self, host='localhost', port=3001,
                 buffer_chunk_size=10**4, buffer_length_limit=10**4):
        self._host = host
        self._port = port
        self._stopping = False
        self.buffer = collections.deque()  # Shared queue of bytes
        self._buffer_chunk_size = buffer_chunk_size   # How many bytes per chunk
        self._buffer_length_limit = buffer_length_limit  # How many chunks in buffer
        self._working = False

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> int:
        return self._port

    @property
    def stopping(self) -> bool:
        return self._stopping

    @property
    def buffer_length_limit(self) -> int:
        return self._buffer_length_limit

    @property
    def buffer_chunk_size(self) -> int:
        return self._buffer_chunk_size

    @property
    def working(self) -> bool:
        return self._working

    async def run_reader(self, reader):
        while not self.stopping:
            try:
                # Stop if buffer is full
                while len(self.buffer) >= self.buffer_length_limit:
                    await asyncio.sleep(1)
                    continue
                try:
                    input_data = await reader.readexactly(self.buffer_chunk_size)
                except asyncio.IncompleteReadError as e:
                    input_data = e.partial
                self.buffer.append(input_data)
            except Exception as e:
                logging.error(e)

    async def run_writer(self, writer):
        while not self.stopping:
            try:
                # Slow down if buffer is short
                if len(self.buffer) < 3:
                    await asyncio.sleep(.1)
                try:
                    input_data = self.buffer.popleft()
                except IndexError:
                    continue
                writer.write(input_data)
                await writer.drain()
            except Exception as e:
                logging.error(e)

    async def forward_bytes(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        self._working = True
        asyncio.ensure_future(self.run_reader(reader=reader))
        asyncio.ensure_future(self.run_writer(writer=writer))

    async def run_server(self):
        reader_server = await asyncio.start_server(client_connected_cb=self.forward_bytes,
                                                   host=self.host, port=self.port)
        async with reader_server:
            await reader_server.serve_forever()
        return

    def stop(self, *_):
        if self.working:
            logging.info("Received interruption signal, stopping...")
            self._stopping = True
        else:
            raise KeyboardInterrupt("Not working yet...")


if __name__ == '__main__':
    log_formatter = logging.Formatter(
        "%(asctime)s [%(module)-15s %(levelname)-8s]     %(message)s",
        style='%'
    )
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_formatter)
    console_handler.setLevel(logging.DEBUG)
    root_logger.addHandler(console_handler)

    loop = asyncio.get_event_loop()
    server = Server(
        host='127.0.0.1',
        port=5000,
    )
    loop.add_signal_handler(signal.SIGINT, server.stop, loop)
    logging.info("Starting file bridging server...")
    loop.run_until_complete(server.run_server())
    loop.close()
    logging.info("Stopped server")
